/* 
   Copyright (C) 1999 Free Software Foundation, Inc.

This file is part of Nilo.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#ifndef __BOOTP_H__
#define __BOOTP_H__

/* This is generated by running configure */
#include "niloconfig.h"

/* Define the basic data types */
#include <sys/types.h>

/* Define the PXE types */
#include "pxe_types.h"

/* standard BOOTP vendor field size */
#define BOOTP_VENDOR  64

#ifdef NILO_USE_EXT_VENDOR
/* extended DHCP vendor field size */
#define BP_VEND_LEN 1024
#else
/* standard DHCP vendor field size */
#ifdef NILO_PXE_COMPLY
#define BP_VEND_LEN  312
#else
#define BP_VEND_LEN  64
#endif
#endif

/* Bootp server IP port*/
#define	BOOTP_SERVER_PORT 67

/* Bootp client IP port */
#define	BOOTP_CLIENT_PORT 68

#define BOOTP_REQUEST	1
#define BOOTP_REPLY	2

#define BP_CHADDR_LEN	16
#define BP_SNAME_LEN	64
#define BP_FILE_LEN	128
#define BP_MINPKTSZ	300	/* to check sizeof(bootp_t) */

/* These definitions come from OSKit's bootp.h file */
#define BOOTP_NET_IP				0x0001
#define BOOTP_NET_NETMASK			0x0002
#define BOOTP_NET_GATEWAY			0x0004
#define BOOTP_NET_SERVER			0x0008
#define BOOTP_NET_DNS_SERVER			0x0010
#define BOOTP_NET_TIME_SERVER			0x0020
#define BOOTP_NET_LOG_SERVER			0x0040
#define BOOTP_NET_LPR_SERVER			0x0080
#define BOOTP_NET_TIME_OFFSET			0x0100
#define BOOTP_NET_HOSTNAME			0x0200
#define BOOTP_NET_SERVER_NAME			0x0400
#define BOOTP_NET_BOOTFILE_NAME			0x0800
#define BOOTP_NET_SERVER_ADDR			0x1000
#define BOOTP_NET_DOMAINNAME			0x2000

#define BOOTP_TIMER_FREQ 100

/* We use unsigned long instead of in_addr for bp_?iaddr because
 * we don't want to drag in netinet/in.h */

/* note that bp_vend can be longer, extending to end of packet. */
struct bootp
{
    u_int8_t bp_opcode;		        /* packet opcode type */
    u_int8_t bp_htype;		        /* hardware addr type */
    u_int8_t bp_hlen;		        /* hardware addr length */
    u_int8_t bp_hops;		        /* gateway hops */
    u_int32_t bp_xid;		        /* transaction ID */
    u_int16_t bp_secs;		        /* seconds since boot began */
    u_int16_t bp_flags;		        /* RFC1532 broadcast, etc. */
    ip4_t bp_ciaddr;		        /* client IP address */
    ip4_t bp_yiaddr;		        /* your IP address */
    ip4_t bp_siaddr;		        /* server IP address */
    ip4_t bp_giaddr;		        /* gateway IP address */
    u_int8_t bp_chaddr[BP_CHADDR_LEN];  /* client hardware address */
    char bp_sname[BP_SNAME_LEN];	/* server host name */
    char bp_file[BP_FILE_LEN];	        /* boot file name */
#ifdef NILO_PXE_COMP
    union {
        u_int8_t d[BP_VEND_LEN];        /* vendor-specific area */
        struct {
            u_int8_t  magic[4];         /* magic number */
            u_int32_t flags;            /* flags/opcodes etc */
            u_int8_t  pad[56];          /* padding chars */
        } v;
    } bp_options;
#else
    /* note that bp_vend can be longer, extending to end of packet. */
    u_int8_t bp_options[BP_VEND_LEN];      /* vendor-specific area */
#endif
};
typedef struct bootp bootp_t;

#define VM_RFC1048      0x63825363L     /* RFC1048 magic number */

/* Tag values used to specify what information is being supplied in
 * the vendor (options) data area of the packet. */
/* RFC 1048 */
#define TAG_END			((u_int8_t) 255)
#define TAG_PAD			((u_int8_t)   0)
#define TAG_SUBNET_MASK		((u_int8_t)   1)
#define TAG_TIME_OFFSET		((u_int8_t)   2)
#define TAG_GATEWAY		((u_int8_t)   3)
#define TAG_TIME_SERVER		((u_int8_t)   4)
#define TAG_NAME_SERVER		((u_int8_t)   5)
#define TAG_DOMAIN_SERVER	((u_int8_t)   6)
#define TAG_LOG_SERVER		((u_int8_t)   7)
#define TAG_COOKIE_SERVER	((u_int8_t)   8)
#define TAG_LPR_SERVER		((u_int8_t)   9)
#define TAG_IMPRESS_SERVER	((u_int8_t)  10)
#define TAG_RLP_SERVER		((u_int8_t)  11)
#define TAG_HOST_NAME		((u_int8_t)  12)
#define TAG_BOOT_SIZE		((u_int8_t)  13)

/* RFC 1395 */
#define TAG_DUMP_FILE		((u_int8_t)  14)
#define TAG_DOMAIN_NAME		((u_int8_t)  15)
#define TAG_SWAP_SERVER		((u_int8_t)  16)
#define TAG_ROOT_PATH		((u_int8_t)  17)

/* RFC 1497 */
#define TAG_EXTEN_FILE		((u_int8_t)  18)

/* RFC 1533 */
#define TAG_NIS_DOMAIN		((u_int8_t)  40)
#define TAG_NIS_SERVER		((u_int8_t)  41)
#define TAG_NTP_SERVER		((u_int8_t)  42)

/* DHCP maximum message size. */
#define TAG_MAX_MSGSZ		((u_int8_t)  57)


extern int fill_bootp_request(bootp_t *bootp_packet,
                              int8_t *macaddr, u_int32_t xid);

extern void dump_bootp_packet(bootp_t *bpr);

#endif /* __BOOTP_H__ */
