/* Nilo - A Network boot loader.

   Copyright (C) 1999 Free Software Foundation, Inc.

This file is part of Nilo.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* This is the include generated by configuring */
#include "niloconfig.h"

/* Define the tracing macros */
#include "trace.h"

/* Define the debug functions that can print the data structures. */
#include "dump.h"

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STRING_H
#include <strings.h>
#endif

/* Define the basic data types */
#include <sys/types.h>

/* Define the PXE types */
#include "pxe_types.h"

/* Define the BOOTP data structures and constants. */
#include "bootp.h"

/* Define the DHCP data structures and constants. */
#include "dhcp.h"

#include <sys/types.h>
#include <sys/time.h>


/* Define the standard BSD TFTP data types */
#include <oskit/freebsd/sys/types.h>
#include <net/if.h>
#include <arpa/tftp.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <arpa/inet.h>

// #include <linux/netdevice.h>
// #include <linux/skbuff.h>
#include <sys/socket.h>
#include "ip.h"
#include "arp.h"
#include "nilo.h"

extern struct protocol_table *prot;

static u_int32_t vend_magic = VM_RFC1048;

/* fill_dhcp_discover -- build a DHCP packet. This is basically a
 *    BOOTP packet with a many new vendor options. In the DHCP case,
 *    it is possible to specify multiple options in the same packet.
 *
 *    params
 *        type - the type of packet. DHCPDISCOVER...DHCP_INFORM.
 *        dhcp_pkt - the allocated memory to build the packet in.
 *        macaddr - the hardware ethernet address of this client.
 *        xid - the transaction ID.
 */
void
fill_dhcp_discover (dhcp_t *pkt)
{
    int i;
    dhcp_opt_msg_type_t *msg_type;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
     packet */
    i = 0;
    while (pkt->bp_options[i++] != TAG_END) {
        msg_type = (dhcp_opt_msg_type_t *)(pkt->bp_options + i);
    }

    msg_type->opt = DHCP_MESSAGE_TYPE;
    msg_type->len = 1;
    msg_type->type = DHCP_DISCOVER;

    /* this packet is only 3 bytes long and sizeof says 4
       cause of padding. We say 2, cause i has already been
       incremented. */
    *(pkt->bp_options + i + 2) = TAG_END;

    /* DHCPDISCOVER packet must always contain the following
       set of options according to PXE. Currently this list is
       based on what a standard IDC DHCPD wants. We'll add full
       PXE support later. */

    /* dhclient uses the following options when talking to dhcpd 3 */
//    add_dhcp_max_msg_len_opt (pkt);
    add_dhcp_request_ip_addr_opt (pkt);
    add_dhcp_hostname_opt (pkt);
    add_dhcp_param_list_opt (pkt);
    add_dhcp_client_id_opt (pkt);

#ifdef NILO_USE_TRACING
    dump_dhcp_packet (pkt);
#endif /* NILO_USE_TRACING */

    NILO_TRACE_FUNCTION_RETURN;
}

void
fill_dhcp_offer (dhcp_t *pkt)
{
    int i;
    dhcp_opt_msg_type_t *msg_type;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = 0;
    while (pkt->bp_options[i++] != TAG_END) {
        msg_type = (dhcp_opt_msg_type_t *)(pkt->bp_options + i);
    }

    /* Add our message type */
    msg_type->opt = DHCP_MESSAGE_TYPE;
    msg_type->len = 1;
    msg_type->type = DHCP_OFFER;

    /* this packet is only 3 bytes long and sizeof says 4
       cause of padding. We say 2, cause i has already been
       incremented. We add the tag now, cause the other options
       look for this to know where to append themselves. */
    *(pkt->bp_options + i + 2) = TAG_END;

    /* DHCPOFFER packet must always contain the following
       set of options according to PXE. Currently this list is
       based on what a standard IDC DHCPD wants. We'll add full
       PXE support later. */

    /* dhclient uses the following options when talking to dhcpd 3 */
    add_dhcp_ipaddr_lease_time_opt (pkt);
    add_dhcp_subnet_mask_opt (pkt);
    add_dhcp_broadcast_addr_opt (pkt);
    add_dhcp_routers_on_subnet_opt (pkt);
    add_dhcp_domainname_opt (pkt);
    add_dhcp_dns_opt (pkt);

#ifdef NILO_USE_TRACING
    dump_dhcp_packet (pkt);
#endif /* NILO_USE_TRACING */

    NILO_TRACE_FUNCTION_RETURN;
}

void
fill_dhcp_request (dhcp_t *pkt)
{
    int i;
    dhcp_opt_msg_type_t *msg_type;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = 0;
    while (pkt->bp_options[i++] != TAG_END) {
        msg_type = (dhcp_opt_msg_type_t *)(pkt->bp_options + i);
    }

    /* Add our message type */
    msg_type->opt = DHCP_MESSAGE_TYPE;
    msg_type->len = 1;
    msg_type->type = DHCP_REQUEST;

    /* this packet is only 3 bytes long and sizeof says 4
       cause of padding. We say 2, cause i has already been
       incremented. We add the tag now, cause the other options
       look for this to know where to append themselves. */
    *(pkt->bp_options + i + 2) = TAG_END;

    /* A DHCPREQUEST packet must always contain the following
       set of options according to PXE. Currently this list is
       based on what a standard IDC DHCPD wants. We'll add full
       PXE support later. */

    /* dhclient uses the following options when talking to dhcpd 3 */
    add_dhcp_server_id_opt (pkt);
    add_dhcp_request_ip_addr_opt (pkt);
    add_dhcp_hostname_opt (pkt);
    add_dhcp_param_list_opt (pkt);
    add_dhcp_client_id_opt (pkt);
 
#ifdef NILO_USE_TRACING
    dump_dhcp_packet (pkt);
#endif /* NILO_USE_TRACING */

    NILO_TRACE_FUNCTION_RETURN;
}

void
fill_dhcp_ack (dhcp_t *pkt)
{
    int i;
    dhcp_opt_msg_type_t *msg_type;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = 0;
    while (pkt->bp_options[i++] != TAG_END) {
        msg_type = (dhcp_opt_msg_type_t *)(pkt->bp_options + i);
    }

    /* Add our message type */
    msg_type->opt = DHCP_MESSAGE_TYPE;
    msg_type->len = 1;
    msg_type->type = DHCP_OFFER;

    /* this packet is only 3 bytes long and sizeof says 4
       cause of padding. We say 2, cause i has already been
       incremented. We add the tag now, cause the other options
       look for this to know where to append themselves. */
    *(pkt->bp_options + i + 2) = TAG_END;

    /* DHCPOFFER packet must always contain the following
       set of options according to PXE. Currently this list is
       based on what a standard IDC DHCPD wants. We'll add full
       PXE support later. */

    /* dhclient uses the following options when talking to dhcpd 3 */
    add_dhcp_server_id_opt (pkt);
    add_dhcp_ipaddr_lease_time_opt (pkt);
    add_dhcp_subnet_mask_opt (pkt);
    add_dhcp_broadcast_addr_opt (pkt);
    add_dhcp_routers_on_subnet_opt (pkt);
    add_dhcp_domainname_opt (pkt);
    add_dhcp_dns_opt (pkt);

#ifdef NILO_USE_TRACING
    dump_dhcp_packet (pkt);
#endif /* NILO_USE_TRACING */

    NILO_TRACE_FUNCTION_RETURN;
}

void
fill_dhcp_header (dhcp_t *pkt, int8_t *macaddr, u_int32_t xid)
{
    NILO_TRACE_FUNCTION;

    /* Build the DHCP packet header */
    memset(pkt, 0, sizeof(*pkt));
    pkt->bp_opcode = BOOTP_REQUEST;
    pkt->bp_htype = 1;
    pkt->bp_hlen = 6;
//    pkt->bp_hops = ;
    pkt->bp_xid = client_info->xid = xid;

    /* this is 4, rather than zero so it'll work for an win32 hosted
       DHCP server */
    pkt->bp_secs = 4;
    pkt->bp_ciaddr = client_info->ciaddr;
    pkt->bp_yiaddr = client_info->yiaddr;
    pkt->bp_siaddr = client_info->siaddr;
    pkt->bp_giaddr = client_info->giaddr;

    memcpy(pkt->bp_chaddr, macaddr, sizeof(pkt->bp_chaddr));
//    pkt->bp_sname = ;
//    pkt->bp_file = ;

    /* Copy in the magic cookie for a BOOTP or DHCP request. */
    vend_magic = htonl(VM_RFC1048);
    memcpy(pkt->bp_options, (u_int8_t *)&vend_magic, sizeof(vend_magic));

    /* Stick the TAG_END, so the other functions can know where
       to start adding their own option. */
    pkt->bp_options[sizeof(vend_magic)] = TAG_END;

    NILO_TRACE_FUNCTION_RETURN;
}

void
add_dhcp_nicif_opt (dhcp_t *pkt)
{
    NILO_TRACE_FUNCTION;

    NILO_TRACE_FUNCTION_RETURN;
}

void
add_dhcp_msg_type_opt (dhcp_t *pkt)
{
    int i;
    dhcp_opt_msg_type_t *msg_type;

    NILO_TRACE_FUNCTION;

    NILO_TRACE_FUNCTION_RETURN;
}

/* dhcp Platform ID opt 97 */
void 
add_dhcp_platformid_opt (dhcp_t *pkt)
{
    int i;
    dhcp_opt_platformid_t *platform_id;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = DHCP_MAX_OPT_SIZE;
    while (pkt->bp_options[i--] != TAG_END) {
        platform_id = (dhcp_opt_platformid_t *)(pkt->bp_options + i);
    }

    NILO_TRACE_PRINTF ("\t\tAdding a Platform ID option #97\n");
    platform_id->opt = DHCP_PLATFORMID;
    platform_id->len = 1+GUID_LEN;
    platform_id->type = DHCP_PLATFORMID_GUID;
    memcpy(platform_id->buf, "156", GUID_LEN); /* FIXME: 156 should be a 16 bit UUID */

    pkt->bp_options[i + 3] = TAG_END;
    NILO_TRACE_FUNCTION_RETURN;
}

/* DHCP PXEClient Class Identifier opt 60 */
void
add_dhcp_class_id_opt (dhcp_t *pkt)
{
    int i, k, j;
    dhcp_opt_class_id_t *class_id;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = DHCP_MAX_OPT_SIZE;
    while (pkt->bp_options[i--] != TAG_END) {
       class_id = (dhcp_opt_class_id_t *)(pkt->bp_options + i);
    }

    class_id->opt = DHCP_CLASS_IDENTIFIER;
    class_id->len = 32;
    for (k = 19; j > 0; --k, j /= 10) {
        class_id->id[k] = (u_int8_t)(class_id->id[k] + j % 10);
    }

    pkt->bp_options[i + 3] = TAG_END;
    NILO_TRACE_FUNCTION_RETURN;
}

/* 9.14. Client-identifier
 *
 * This option is used by DHCP clients to specify their unique
 * identifier. DHCP servers use this value to index their database of address
 * bindings. This value is expected to be unique for all clients in an
 * administrative domain.  Identifiers SHOULD be treated as opaque objects by
 * DHCP servers.  The client identifier MAY consist of type-value pairs
 * similar to the 'htype'/'chaddr' fields defined in [3]. For instance, it MAY
 * consist of a hardware type and hardware address. In this case the type
 * field SHOULD be one of the ARP hardware types defined in STD2 [22]. A
 * hardware type of 0 (zero) should be used when the value field contains an
 * identifier other than a hardware address (e.g. a fully qualified domain
 * name).  For correct identification of clients, each client's client-
 * identifier MUST be unique among the client-identifiers used on the subnet
 * to which the client is attached. Vendors and system administrators are
 * responsible for choosing client-identifiers that meet this requirement for
 * uniqueness.
 *
 * The code for this option is 61, and its minimum length is 2.
 */
void
add_dhcp_client_id_opt (dhcp_t *pkt)
{
    int i;
    dhcp_opt_client_id_t *client_uuid;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = DHCP_MAX_OPT_SIZE;
    while (pkt->bp_options[i--] != TAG_END) {
        client_uuid = (dhcp_opt_client_id_t *)(pkt->bp_options + i);
    }

    pkt->bp_options[++i] = DHCP_CLIENT_IDENTIFIER;
    pkt->bp_options[++i] = 4;
    memcpy((pkt->bp_options + ++i), "nilo", 4);
    pkt->bp_options[i+4] = TAG_END;

//    client_uuid->opt = DHCP_CLIENT_IDENTIFIER;
//    client_uuid->len = 4;
//    client_uuid->id = 0;                    /* FIXME: bogus value */

//    *(pkt->bp_options + i + 3) = TAG_END;

    NILO_TRACE_FUNCTION_RETURN;
}

/* 9.10. Maximum DHCP Message Size
 *
 * This option specifies the maximum length DHCP message that it is willing to
 * accept. The length is specified as an unsigned 16-bit integer. A client may
 * use the maximum DHCP message size option in DHCPDISCOVER or DHCPREQUEST
 * messages, but should not use the option in DHCPDECLINE messages. 
 *
 * The code for this option is 57, and its length is 2. The minimum legal
 * value is 576 octets.
 */
void
add_dhcp_max_msg_len_opt (dhcp_t *pkt)
{
    int i;
    dhcp_opt_max_msg_size_t *msg_len;

    NILO_TRACE_FUNCTION;

    i = DHCP_MAX_OPT_SIZE;
    while (pkt->bp_options[i--] != TAG_END) {
        msg_len = (dhcp_opt_max_msg_size_t *)(pkt->bp_options + i);
    }

    NILO_TRACE_PRINTF ("\t\tAdding a Network Type option #57\n");
    msg_len->opt = DHCP_MAXMSGSIZE;
    msg_len->len = 2;
    msg_len->size = INTSWAP(sizeof(dhcp_t));

    /* this packet is only 3 bytes long and sizeof says 4
       cause of padding. We say 2, cause i has already been
       incremented. */
    *(pkt->bp_options + i + 3) = TAG_END;

    NILO_TRACE_FUNCTION_RETURN;
}

/* DHCP Client System Architecture opt 93 */
void
add_dhcp_sysarch_opt (dhcp_t *pkt)
{
    int i;
    dhcp_opt_sysarch_t *client_sysarch;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = DHCP_MAX_OPT_SIZE;
    while (pkt->bp_options[i--] != TAG_END) {
        client_sysarch = (dhcp_opt_sysarch_t *)(pkt->bp_options + i);
    }

    NILO_TRACE_PRINTF ("\t\tAdding a System Architecture option #93\n");
    client_sysarch->opt = DHCP_SYSARCH;
    client_sysarch->len = 2;
    client_sysarch->type = DHCP_STD_X86;    /* FIXME: are we sure ? */

    *(pkt->bp_options + i + 3) = TAG_END;
    NILO_TRACE_FUNCTION_RETURN;
}

/* 9.8. Parameter Request List
 *
 * This option is used by a DHCP client to request values for specified
 * configuration parameters.  The list of requested parameters is specified as
 * n octets, where each octet is a valid DHCP option code as defined in this
 * document.  The client MAY list the options in order of preference. The DHCP
 * server is not required to return the options in the requested order, but
 * MUST try to insert the requested options in the order requested by the
 * client.
 *
 * The code for this option is 55. Its minimum length is 1.
 */
void
add_dhcp_param_list_opt (dhcp_t *pkt)
{
    int i, j;
    dhcp_opt_param_list_t *client_param;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = DHCP_MAX_OPT_SIZE;
    while (pkt->bp_options[i--] != TAG_END) {
       client_param = (dhcp_opt_param_list_t *)(pkt->bp_options + i);
    }

    NILO_TRACE_PRINTF ("\t\tAdding a Parameter Request option #55\n");
    client_param->opt = DHCP_PARAM_REQUEST;
    client_param->len = 13;
    j = 0;
    client_param->codes[j++] = DHCP_SUBNET_MASK;
    client_param->codes[j++] = DHCP_ROUTERS_ON_SUBNET;
    client_param->codes[j++] = DHCP_CLASS_IDENTIFIER;
    client_param->codes[j++] = DHCP_VENDOR_SPECIFICINFO;
    client_param->codes[j++] = DHCP_BOOTFILE;
    client_param->codes[j++] = 0x80;
    client_param->codes[j++] = 0x81;
    client_param->codes[j++] = 0x82;
    client_param->codes[j++] = 0x83;
    client_param->codes[j++] = 0x84;
    client_param->codes[j++] = 0x85;
    client_param->codes[j++] = 0x86;
    client_param->codes[j++] = 0x87;

    /* Add the end tag */
    *(pkt->bp_options + i + j + 3) = TAG_END;

    NILO_TRACE_FUNCTION_RETURN;
}

/* DHCP Network Type opt 94 */
void
add_dhcp_nic_type_opt (dhcp_t *pkt)
{
    int i;
    dhcp_opt_nicif_t *client_net_type;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = DHCP_MAX_OPT_SIZE;
    while (pkt->bp_options[i--] != TAG_END) {
        client_net_type = (dhcp_opt_nicif_t *)(pkt->bp_options + i);
    }

    NILO_TRACE_PRINTF ("\t\tAdding a Network Type option #94\n");
    client_net_type->opt = DHCP_NICIF;
    client_net_type->len = 3;
    client_net_type->type = DHCP_UNDI;  /* Supports a UNDI API. Other
                                           values are DHCP_PCI
                                           and DHCP_PNP */
    client_net_type->major = 2;         /* PXE spec major number */
    client_net_type->minor = 1;         /* PXE spec minor number */

    *(pkt->bp_options + i + 3) = TAG_END;

    NILO_TRACE_FUNCTION_RETURN;
}

/* section 9.1. Requested IP Address
 *
 * This option is used in a client request (DHCPDISCOVER) to allow the
 * client to request that a particular IP address be assigned. 
 *
 * The code for this option is 50, and its length is 4.
 */
void
add_dhcp_request_ip_addr_opt (dhcp_t *pkt)
{
    int i, j;
    u_int32_t num;
    dhcp_opt_request_ip_t *ip_addr;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = DHCP_MAX_OPT_SIZE;
    while (pkt->bp_options[i--] != TAG_END) {
        ip_addr = (dhcp_opt_request_ip_t *)(pkt->bp_options + i);
    }

    NILO_TRACE_PRINTF ("\t\tAdding a Request IP Address option #50\n");
    pkt->bp_options[++i] = DHCP_REQUESTED_IP_ADDR;
    pkt->bp_options[++i] = 4;
    num = client_info->yiaddr.num;
//    num = ntohl(client_info->yiaddr);
    memcpy((pkt->bp_options + ++i), &num, 4);

//    ip_addr->opt = DHCP_REQUESTED_IP_ADDR;
//    ip_addr->len = 4;
    /* FIXME: This IP number shouldn't be hardcoded */
//    ip_addr->ip.num = 0xc0cbbc0f;       /* 192.203.188.11 */

    pkt->bp_options[i+4] = TAG_END;

    NILO_TRACE_FUNCTION_RETURN;
}

/* section 3.14. Host Name Option
 *
 * This option specifies the name of the client. The name may or may not be
 * qualified with the local domain name (see section 3.17 for the preferred
 * way to retrieve the domain name). See RFC 1035 for character set
 * restrictions.
 *
 * The code for this option is 12, and its minimum length is 1. 
 */
void
add_dhcp_hostname_opt (dhcp_t *pkt)
{
    int i;
    dhcp_opt_hostname_t *name;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = DHCP_MAX_OPT_SIZE;
    while (pkt->bp_options[i--] != TAG_END) {
        name = (dhcp_opt_hostname_t *)(pkt->bp_options + i);
    }

    NILO_TRACE_PRINTF ("\t\tAdding a Hostname option #12\n");

    pkt->bp_options[++i] = DHCP_HOSTNAME;
    pkt->bp_options[++i] = 4;
    memcpy((pkt->bp_options + ++i),  "nilo", 4);
    pkt->bp_options[i+4] = TAG_END;

//    name->opt = DHCP_HOSTNAME;
//    name->len = 4;
//    memcpy (name->filename, "nilo\0", 5);

//    *(pkt->bp_options + i + 7) = TAG_END;
    NILO_TRACE_FUNCTION_RETURN;
}

/* These are the options commonly used in a DHCPOFFER mesage */

/* section 9.2. IP Address Lease Time
 *
 * This option is used in a client request (DHCPDISCOVER or DHCPREQUEST) to
 * allow the client to request a lease time for the IP address. In a server
 * reply (DHCPOFFER), a DHCP server uses this option to specify the lease time
 * it is willing to offer.  
 * The time is in units of seconds, and is specified as a 32-bit unsigned
 * integer.  
 *
 * The code for this option is 51, and its length is 4.
 */
void
add_dhcp_ipaddr_lease_time_opt (dhcp_t *pkt)
{
    int i;
    dhcp_opt_lease_time_t *lease;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = DHCP_MAX_OPT_SIZE;
    while (pkt->bp_options[i--] != TAG_END) {
        lease = (dhcp_opt_lease_time_t *)(pkt->bp_options + i);
    }

    NILO_TRACE_PRINTF ("\t\tAdding a IP Address Lease Time option #51\n");

    lease->opt = DHCP_IPADDR_LEASE_TIME;
    lease->len = 4;
    lease->time = client_info->leasetime;

    *(pkt->bp_options + i + 7) = TAG_END;
    NILO_TRACE_FUNCTION_RETURN;
}

/* section 3.3 Subnet Mask
 *
 * The subnet mask option specifies the client's subnet mask as per
 * RFC 950 [5]. If both the subnet mask and the router option are
 * specified in a DHCP reply, the subnet mask option MUST be first. 
 *
 * The code for the subnet mask option is 1, and its length is 4 octets. */
void
add_dhcp_subnet_mask_opt (dhcp_t *pkt)
{
    int i;
    u_int32_t num = 0xffffff00;
    dhcp_opt_subnet_mask_t *mask;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = DHCP_MAX_OPT_SIZE;
    while (pkt->bp_options[i--] != TAG_END) {
       mask = (dhcp_opt_subnet_mask_t *)(pkt->bp_options + i);
    }

    NILO_TRACE_PRINTF ("\t\tAdding a Subnet Mask option #1\n");
    mask->opt = DHCP_SUBNET_MASK;
    mask->len = 4;
//    memcpy((pkt->bp_options + i + 3), &num, 4);
    mask->mask.num = client_info->mask.num;

    /* add the end tag */
    *(pkt->bp_options + i + 7) = TAG_END;
    NILO_TRACE_FUNCTION_RETURN;
}

/* section 5.3. Broadcast Address Option
 *
 * This option specifies the broadcast address in use on the client's
 * subnet. Legal values for broadcast addresses are specified in section
 * 3.2.1.3 of [4]. 
 *
 * The code for this option is 28, and its length is 4.
 */
void
add_dhcp_broadcast_addr_opt (dhcp_t *pkt)
{
    int i;
    dhcp_opt_broadcast_addr_t *addr;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = DHCP_MAX_OPT_SIZE;
    while (pkt->bp_options[i--] != TAG_END) {
        addr = (dhcp_opt_broadcast_addr_t *)(pkt->bp_options + i);
    }
    NILO_TRACE_PRINTF ("\t\tAdding a Broadcast Address option #28\n");

    addr->opt = DHCP_BROADCAST_ADDR;
    addr->len = 4;
    addr->addr.num = 0xffff00;          /* FIXME: this doesn't get written right if it's 32 bit*/

    /* add the end tag */
    *(pkt->bp_options + i + 7) = TAG_END;
    NILO_TRACE_FUNCTION_RETURN;
}

/* section 3.5 Router Option
 *
 * The router option specifies a list of IP addresses for routers on the
 * client's subnet. Routers SHOULD be listed in order of preference. 
 *
 * The code for the router option is 3. The minimum length for the router
 * option is 4 octets, and the length MUST always be a multiple of 4.
 */
void
add_dhcp_routers_on_subnet_opt (dhcp_t *pkt)
{
    int i;
    dhcp_opt_router_t *list;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = DHCP_MAX_OPT_SIZE;
    while (pkt->bp_options[i--] != TAG_END) {
        list = (dhcp_opt_router_t *)(pkt->bp_options + i);
    }
    NILO_TRACE_PRINTF ("\t\tAdding a Routers On Subnet option #3\n");

    list->opt = DHCP_ROUTERS_ON_SUBNET;
    list->len = 4;
//    list-> = ;

    /* add the end tag */
    *(pkt->bp_options + i + 7) = TAG_END;
    NILO_TRACE_FUNCTION_RETURN;
}

/* section 3.17. Domain Name
 *
 * This option specifies the domain name that client should use when
 * resolving hostnames via the Domain Name System. 
 *
 * The code for this option is 15. Its minimum length is 1. 
 */
void
add_dhcp_domainname_opt (dhcp_t *pkt)
{
    int i;
    dhcp_opt_domainname_t *name;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = DHCP_MAX_OPT_SIZE;
    while (pkt->bp_options[i--] != TAG_END) {
        name = (dhcp_opt_domainname_t *)(pkt->bp_options + i);
    }
    NILO_TRACE_PRINTF ("\t\tAdding a Domain Name option #15\n");

    name->opt = DHCP_DOMAINNAME;
    name->len = 4;

    /* add the end tag */
    *(pkt->bp_options + i + 7) = TAG_END;
    NILO_TRACE_FUNCTION_RETURN;
}

/* section 3.8. Domain Name Server Option
 *
 * The domain name server option specifies a list of Domain Name System
 * (STD 13, RFC 1035 [8]) name servers available to the client. Servers
 * SHOULD be listed in order of preference.
 *
 * The code for the domain name server option is 6. The minimum length for
 * this option is 4 octets, and the length MUST always be a multiple of 4. 
 */
void
add_dhcp_dns_opt (dhcp_t *pkt)
{
    int i;
    dhcp_opt_dns_t *name;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = DHCP_MAX_OPT_SIZE;
    while (pkt->bp_options[i--] != TAG_END) {
        name = (dhcp_opt_dns_t *)(pkt->bp_options + i);
    }
    NILO_TRACE_PRINTF ("\t\tAdding a DNS option #6\n");

    name->opt = DHCP_DNS;
    name->len = 4;

    /* add the end tag */
    *(pkt->bp_options + i + 7) = TAG_END;
    NILO_TRACE_FUNCTION_RETURN;
}

/* 9.7. Server Identifier
 *
 * This option is used in DHCPOFFER and DHCPREQUEST messages, and may
 * optionally be included in the DHCPACK and DHCPNAK messages. DHCP servers
 * include this option in the DHCPOFFER in order to allow the client to
 * distinguish between lease offers. DHCP clients use the contents of the
 * 'server identifier' field as the destination address for any DHCP messages
 * unicast to the DHCP server. DHCP clients also indicate which of several
 * lease offers is being accepted by including this option in a DHCPREQUEST
 * message.  The identifier is the IP address of the selected server.
 *
 * The code for this option is 54, and its length is 4.
 */
void
add_dhcp_server_id_opt (dhcp_t *pkt)
{
    int i;
    dhcp_opt_server_id_t *sid;

    NILO_TRACE_FUNCTION;

    /* Scan the packet for TAG_END, and add ourselves right after
     the previous option, and move the TAG_end to the end of the
    packet */
    i = DHCP_MAX_OPT_SIZE;
    while (pkt->bp_options[i--] != TAG_END) {
        sid = (dhcp_opt_server_id_t *)(pkt->bp_options + i);
    }
    NILO_TRACE_PRINTF ("\t\tAdding a Server Identifier option #54\n");

    sid->opt = DHCP_SERVER_IDENTIFIER;
    sid->len = 4;

    /* add the end tag */
    *(pkt->bp_options + i + 7) = TAG_END;
    NILO_TRACE_FUNCTION_RETURN;
}




