/* 
   Copyright (C) 1999 Free Software Foundation, Inc.

This file is part of Nilo.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#ifndef __PXE_TFTP_H__
#define __PXE_TFTP_H__

/* This is generated by running configure */
#include "niloconfig.h"

/* Define the basic data types */
#include <sys/types.h>

/* Define the PXE data types */
#include "pxe_types.h"

/* NOTE: None of the TFTP API services are available after the stack
 *	 has been unloaded.
 */

/* Constants that define the offsets into the table of functions.
 * One of these must be loaded into the BX register before making
 * a call a TFTP API service. 
 */
#define PXENV_TFTP_OPEN			0x20
#define PXENV_TFTP_CLOSE		0x21
#define PXENV_TFTP_READ			0x22
#define PXENV_TFTP_READ_FILE		0x23
#define PXENV_TFTP_GET_FSIZE		0x25

/* Data structure for pxenv_tftp_open().
 * parameters:
 *     server_ip - the IP address of the TFTP server in Big Endian
 *                 format.
 *     gw_ip - The IP address of gateway in Big Endian
 *             format.
 *     filename - a NULL terminated string represnting the boot
 *             file name.
 *     tftpport - the port the TFTP server is listening on. This must
 *                be in Big Endian format.
 *     packetsize - the requested size of the TFTP packets. This must
 *                  be over 512 bytes.
 * returns:
 *     status - The exit status of this function. This is set to one
 *	        of the PXENV_STATUS_xxx constants.
 *     packetsize - the negotiated size of the TFTP packets. Must be
 *                  at least 512 bytes.
 */
typedef struct pxenv_tftp_open_s {
    pxenv_status_t status;
    ip4_t server_ip;
    ip4_t gw_ip;
    u_int8_t filename[128];
    tftp_port_t tftpport;
    int16_t packetsize;
} pxenv_tftp_open_t;

/* Data structure for pxenv_tftp_get_fsize().
 * parameters:
 *     server_ip - the IP address of the TFTP server in Big Endian
 *                 format.
 *     gw_ip - The IP address of gateway in Big Endian
 *             format.
 *     filename - a NULL terminated string represnting the boot
 *             file name.
 *     tftpport - the port the TFTP server is listening on. This must
 *                be in Big Endian format.
 * returns:
 *     status - The exit status of this function. This is set to one
 *	        of the PXENV_STATUS_xxx constants.
 *     filesize - the size of the boot file in bytes.
 */
typedef struct pxenv_tftp_get_fsize_s {
    pxenv_status_t status;
    ip4_t server_ip;
    ip4_t gw_ip;
    u_int8_t filename[128];
    u_int32_t filesize;
} pxenv_tftp_get_fsize_t;

/* Data structure for pxenv_tftp_close().
 * parameters:
 *     none
 * returns:
 *     status - The exit status of this function. This is set to one
 *	        of the PXENV_STATUS_xxx constants.
 */
typedef struct pxenv_tftp_close_s {
    pxenv_status_t status;
} pxenv_tftp_close_t;

/* Data structure for pxenv_tftp_read().
 * parameters:
 *     buffersize - the size of the receive buffer in bytes. This must
 *                  be larger than 512.
 *     buffer - the buffer address in offset:segment format.
 * returns:
 *     status - The exit status of this function. This is set to one
 *	        of the PXENV_STATUS_xxx constants.
 *     packetnum - the packet sequence number.
 *     buffersize - the packet size in bytes. This must be less than
 *                  the requested buffer size.
 */
typedef struct pxenv_tftp_read_s {
    pxenv_status_t status;
    u_int16_t packetnum;
    u_int16_t buffersize;
    segoff16_t Buffer;
} pxenv_tftp_read_t;

/* Data structure for pxenv_tftp_read_file().
 * parameters:
 *     filename - a NULL terminated string represnting the boot
 *             file name.
 *     buffersize - the size of the receive buffer in bytes. 
 *     buffer - the buffer address in offset:segment format.
 *     server_ip - the IP address of the TFTP server in Big Endian
 *                 format.
 *     gw_ip - the IP address of gateway in Big Endian
 *             format.
 *     mcast_ip - the multicast IP address in Big Endian format.
 *     tftp_clnt_port - the port the client listens to for a multicast.
 *     tftp_srv_port - the port the server listens to for a multicast.
 *     tftp_open_timeout - the timeout value in seconds used for
 *                         receiving data. If zero, the default of
 *			   TFTP timeout is used.
 *     tftp_reopen_delay - the maximum wait delay time in seconds
 *                         between ACK of a fast packet and a new
 *                         TFTP open.
 * returns:
 *     status - The exit status of this function. This is set to one
 *	        of the PXENV_STATUS_xxx constants.
 *     packetnum - the packet sequence number.
 *     buffersize - the size of the packet received in the buffer.
 */
typedef struct pxenv_tftp_read_file_s {
    pxenv_status_t status;
    u_int8_t filename[128];
    u_int32_t buffersize;
    addr32_t buffer;
    ip4_t server_ip;
    ip4_t gw_ip;
    ip4_t mcast_ip;
    udp_port_t tftp_clnt_port;
    udp_port_t tftp_srv_port;
    u_int16_t tftp_open_timeout;
    u_int16_t tftp_reopen_delay;
} pxenv_tftp_read_file_t;

/* Note:
 * If the McastIPAddress specifies a non-zero value, the
 * pxenv_tftp_read_file() call tries to listen for multicast packets
 * on the tftp_clnt_port before opening a TFTP/MTFTP connection to the
 * server.  If it receives any packets (and not all) or if does not
 * receive any, it waits for specified time and tries to reopen a
 * multicast connection to the server.  If the server supports
 * multicast, it notifies the acknowledging client with a unicast and
 * starts sending (multicast) the file.  If the multicast open request
 * times out, the client tries to connect to the server at TFTP server
 * port for a unicast transfer.
 */

/* prototypes for the PXE TFTP API functions */
pxenv_ret_code_t
pxenv_tftp_open (pxenv_tftp_open_t tftpopen);
pxenv_ret_code_t
pxenv_tftp_close (pxenv_tftp_close_t tftpclose);
pxenv_ret_code_t
pxenv_tftp_read (pxenv_tftp_read_t tftpread);
pxenv_ret_code_t
pxenv_tftp_read_file (pxenv_tftp_read_file_t tftpfile);
pxenv_ret_code_t
pxenv_tftp_get_fsize (pxenv_tftp_get_fsize_t tftpfsize);

#endif /* __PXE_TFTP_H__ */



