/*
 * Copyright (c) 1997-1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/* Modified from FreeBSD 2.1 sys/i386/boot/netboot */
/**************************************************************************
NETBOOT -  BOOTP/TFTP Bootstrap Program

Author: Martin Renters
  Date: Dec/93

Modified by: Rob Savoye <rob@welcomehome.org>
  Date: April/99
**************************************************************************/

/* This is the include generated by configuring */
#include "niloconfig.h"

#include <oskit/c/sys/types.h>
#include <oskit/net/ether.h>
#include <oskit/c/arpa/inet.h>
#include <oskit/c/string.h>
#include <oskit/c/assert.h>
#include <oskit/c/stdio.h>

/* Define the PXE types */
#include "pxe_types.h"

/* Define the tracing macros */
#include "trace.h"

#include "udp.h"
#include "arp.h"
#include "tftp.h"
#include "dhcp.h"

/* The arptable is maintained in host order. */
//arptable_t arptable[MAX_ARP];

static char ether_bcast[] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

/*
 * IP checksum algorithm.
 */
u_int16_t
//dhcp_ipchksum(struct iphdr *ipv, int len)
dhcp_ipchksum(void *ipv, int len)
{
    u_int16_t *ip = (u_int16_t *)ipv;
    u_int32_t sum = 0;
    
    NILO_TRACE_FUNCTION;
    
    len >>= 1;
    while (len--) {
        sum += *(ip++);
        if (sum > 0xFFFF)
            sum -= 0xFFFF;
    }

    NILO_TRACE_FUNCTION_RETURN;

    return((~sum) & 0x0000FFFF);
}


int32_t
dhcp_udp_broadcast(char *ether_src, 
	unsigned long srcip,
	unsigned short srcsock, 
	unsigned short destsock, int len, void *bufv)
{
    char *buf = bufv;
    struct iphdr *ip;
    struct udphdr *udp;
    struct udp_pkt *full_pkt;
    u_int32_t temp;
    
    NILO_TRACE_FUNCTION;

    if ((full_pkt = (struct udp_pkt *)malloc(sizeof(struct udp_pkt))) == 0) {
        panic("Called malloc any memory for full_pkt");
    }
    
    len += sizeof(struct iphdr) + sizeof(struct udphdr);
    memcpy (&full_pkt->u.dhcp, bufv, sizeof(dhcp_t));
    full_pkt->ip.verhdrlen = 0x45;
    full_pkt->ip.service = 0;
    full_pkt->ip.len = htons(len);
    full_pkt->ip.ident = 0;
    full_pkt->ip.frags = 0;
    full_pkt->ip.ttl = 60;
    full_pkt->ip.protocol = IP_UDP;
    full_pkt->ip.chksum = 0;
    
    temp = srcip;
    memcpy(full_pkt->ip.src, &temp, 4);
    temp = IP_BROADCAST;
    memcpy(full_pkt->ip.dest, &temp, 4);
    
    full_pkt->ip.chksum = dhcp_ipchksum(&full_pkt->ip, sizeof(struct iphdr));
    full_pkt->udp.src = htons(srcsock);
    full_pkt->udp.dest = htons(destsock);
    full_pkt->udp.len = htons(len - sizeof(struct iphdr));
    full_pkt->udp.chksum = 0;

    dhcp_eth_transmit(ether_src, ether_bcast, ETHERTYPE_IP, len, full_pkt);

    NILO_TRACE_FUNCTION_RETURN;

    return(1);
}


int32_t
tftp_udp_transmit(char *ether_src, u_int32_t srcip,
        u_int32_t destip, u_int32_t netmask, u_int16_t srcsock, 
	u_int16_t destsock, int32_t len, void *bufv)
{
    char *buf = bufv;
    struct iphdr *ip;
    struct udphdr *udp;
    struct udp_pkt *full_pkt;
    struct arprequest arpreq;
    int arpentry, i;
    int retry = 6;
    u_int32_t temp;
 
    NILO_TRACE_FUNCTION;

    if ((full_pkt = (struct udp_pkt *)malloc(sizeof(struct udp_pkt))) == 0) {
        panic("Called malloc any memory for full_pkt");
    }
    
    len += sizeof(struct iphdr) + sizeof(struct udphdr);
    memcpy (&full_pkt->u.tftp, bufv, sizeof(tftp_t));
    full_pkt->ip.verhdrlen = 0x45;      /* int8_t */
    full_pkt->ip.service = 0;           /* int8_t */
    full_pkt->ip.len = htons(len);      /* u_int16_t */
    /* FIXME: Ident should be random */
    full_pkt->ip.ident = 0;             /* u_int16_t */
    full_pkt->ip.frags = 0;             /* u_int16_t */
    full_pkt->ip.ttl = 64;              /* int8_t was 60 */
    full_pkt->ip.protocol = IP_UDP;     /* int8_t */
    full_pkt->ip.chksum = 0;
    temp = srcip;
    memcpy(full_pkt->ip.src, &temp, 4);
    temp = destip;
    memcpy(full_pkt->ip.dest, &temp, 4);
    full_pkt->ip.chksum = dhcp_ipchksum(&full_pkt->ip,
                                        sizeof(struct iphdr));
    
    full_pkt->udp.src = htons(srcsock);
    full_pkt->udp.dest = htons(destsock);
    full_pkt->udp.len = htons(len - sizeof(struct iphdr));
    full_pkt->udp.chksum = 0;           /* u_int16_t */
//    full_pkt->udp.chksum = dhcp_ipchksum(&full_pkt->udp, sizeof(struct udphdr));

#ifdef NILO_USE_TRACINGXXX
    dump_ip_packet (&full_pkt->ip);
    dump_udp_packet (&full_pkt->udp);
#endif  /* NILO_USE_TRACING  */

//    memcpy (&full_pkt->u.tftp, bufv, sizeof(tftp_t));
    dhcp_eth_transmit(ether_src, full_pkt->ip.dest, ETHERTYPE_IP,
                      len, full_pkt);

//    dhcp_eth_transmit(ether_src, ether_bcast, ETHERTYPE_IP, len, full_pkt);

    NILO_TRACE_FUNCTION_RETURN;

    return(1);
}
